#include "Quadric.h"

#include <algorithm>

Quadric::Quadric(Mat4f pQ, Shader pShader){
    Q = pQ;
    shader = pShader;
}

Intersection Quadric::intersect(Ray ray, float tStart, float tEnd){
    // Initialize result struct (per default, hasIntersection is false):
    Intersection intersection;

    // Define how many steps should be made for linear search:
    unsigned int linearSearchSteps = std::min((unsigned int)(ray.direction.length() * (tEnd - tStart) * 20), 100u);

    // In this variable we store the ray parameter found by the linear search:
    float linearSearchT = 0;

    float linearSearchStepSize = (1/float(linearSearchSteps)) * (tEnd - tStart);

    /*
     * We first perform a rough linear search to find the approximate intersection
     * with the metaball, which is refined by the later interval bisection that
     * you should implement.
     *
     * We do not apply interval bisection directly, since it does not guarantee
     * that the first zero will be found.
     */

    {
        // Stores if an intersection was found:
        bool foundIntersection = false;

        // Get the potential field value at X:
        float f=0;

        // Iterate along the ray (step-by-step):
        for(unsigned int i=0; i < linearSearchSteps; ++i){
            // Calculate the parameter t:
            float t = tStart + i * linearSearchStepSize;

            // Get the position of the ray at parameter t:
            Vec4f X = ray.evaluate(t);

            // Get the potential field value at X:
            f = F(X);

            // When the surface gets near enough to 0:
            if(std::abs(f) < 0.1f){
                foundIntersection = true;
                break;
            }

            // Update the search result:
            linearSearchT = t;
        }

        // If the last potential field is smaller 0, no surface was hit and
        // we return an empty Intersection object where hasIntersection is false:
        if(!foundIntersection)
            return intersection;
    }

    // When an approximate intersection with the metaball was found, we redefine
    // the range in which the interval bisection should search again:
    tStart = linearSearchT - linearSearchStepSize;
    tEnd = linearSearchT + linearSearchStepSize;

    // Use this EPSILON to test whether we are finally near enough to 0:
    //float EPSILON = 0.00001f;

    float EPSILON = 0.0001f;
    // Perform interval bisection:
    for(unsigned int i=0; i<16; ++i){
        /*
         * TODO (Assignment Sheet 2, Exercise 1b):
         *
         * Implement an interval bisection in the range [tStart, tEnd] to refine the
         * found intersection. In each iteration step, the range should be halved.
         * You don't have to modify any other variable than tStart and tEnd.
         *
         * Use the 'EPSILON' variable to test whether we are finally near enough to 0
         * to stop the iteration.
         *
         * There are three cases that you have to consider:
         *  1) The middle point is on the isosurface.
         *  2) The middle point is inside the meta ball.
         *  3) The middle point is outside the meta ball.
         */

        // compute the sum for the middle point
        float t12 = (tStart+tEnd)/2.f;

        Vec4f x = ray.evaluate(t12);
        float sum12 = F(x);

        // middle point is already on the isosurface
        if(std::abs(sum12 - 0.1f) < EPSILON )
        {
            tStart = tEnd = t12;
            break;
        }
        else
        {
            // compute the sum for the other points
            x = ray.evaluate(tStart);
            float sum0 = std::abs(F(x) - 0.1f);

            x = ray.evaluate(tEnd);
            float sum1 = std::abs(F(x) - 0.1f);

            if( sum0 < sum1 )
                tEnd = t12;
            else
                tStart = t12;
        }

    }

    // Use the mid point of [tStart, tEnd] as parameter for the intersection:
    float t = (tStart+tEnd)/2.f;

    // Calculate the intersection candidate:
    Vec4f x = ray.evaluate(t);

    // Evaluate the potential field at the intersection candidate:
    float fx = F(x);

    // If the intersection point is not near enough to 0, return an empty
    // intersection (where hasIntersection is false):
    if(std::abs(fx) > 0.2f)
        return intersection;

    // Calculate the normal:
    float eps = 0.0001f;
    Vec4f n = -Vec4f(
        F(Vec4f(x[0]-eps, x[1], x[2])) -  F(Vec4f(x[0]+eps, x[1], x[2])),
        F(Vec4f(x[0], x[1]-eps, x[2])) -  F(Vec4f(x[0], x[1]+eps, x[2])),
        F(Vec4f(x[0], x[1], x[2]-eps)) -  F(Vec4f(x[0], x[1], x[2]+eps)),
        0.f
    ).normalized();

    intersection.hasIntersection = true;
    intersection.t = t;
    intersection.hitPoint = ray.evaluate(t);
    intersection.normal = n.normalized();
    intersection.shader = &shader;

    return intersection;
}

float Quadric::F(Vec4f X){
    /*
     * TODO (Exercise 1a)
     *
     *
     */

    Vec4f qX = (Q * X);

    return qX.dot(X);
}
